#include <math.h>
#include "galias.h"
#include "alias.h"
#include "effstress.h"
#include "matrix.h"
#include "vector.h"
#include "global.h"
#include "intpoints.h"
#include "vecttens.h"
#include "iotools.h"



effstress::effstress()
{
  ppv = 0.0;
}



effstress::~effstress()
{
}



/**
  The function reads type of pore pressure computation and in the 
  case of contant for all problem solved, the value is read from 
  the input file.
 
  @param in - pointer to the opened XFILE

  @return The funtion does not return anything.

  Created by Tomas Koudelka, 2.10.2013
*/
void effstress::read(XFILE *in)
{
  xfscanf(in, "%k%m", "pore_press_comp", &pore_press_comp_kwdset, &ppct);
  switch (ppct)
  {
    case noporep:
      break;
    case const_all:
      xfscanf(in, "%le", &ppv);
      if (Mp->pore_press < 2)
        Mp->pore_press = 1;
      else
      {
        print_err("partially coupled and fully coupled approaches cannot be used together", __FILE__, __LINE__, __func__);
        abort();
      }
      break;
    case nonmechq:
      if (Mp->pore_press == 1)
      {
        print_err("partially coupled and fully coupled approaches cannot be used together", __FILE__, __LINE__, __func__);
        abort();
      }
      else
      {
        // partially coupled approach is supposed
        // it can be rewritten in METR according to METR input file
        Mp->pore_press = 2; 
      }
      break;
    default:
      print_err("Unknown type of pore pressure computation is required", __FILE__, __LINE__, __func__);
  }

  return;
}



/**
  The function calculates stresses according to effective stress concept.

  @param ipp - number of integration point
  @param im  - index of the material in the ipp tm and idm arrays. The standard value is zero.
  @param ido - index of internal variables in the ip's ipp eqother array. The standard value is zero.

  @return The function does not return anything but fills stress array on the given integration point  
          with the actual stress values.

  Created by TKo & TKr, 09.2012
  Modified by TKo, 10.2013
*/
void effstress::nlstresses (long ipp, long im, long ido)
{
  long i;
  long ncomp = Mm->ip[ipp].ncompstr;
  vector sig(ncomp);
  matrix sigt(3,3);
  double pp;
  
  pp = Mm->givenonmechq(pore_pressure,ipp);    // actual pore pressure
  
  // compute effective stresses
  Mm->computenlstresses(ipp, im+1, ido);
  // retrieve effective stress vector
  Mm->givestress(0, ipp, sig);
  
  // create tensor form of effective stress vector - sigt
  vector_tensor(sig, sigt, Mm->ip[ipp].ssst,stress);
  
  // calculate total stress tensor
  // sig^{tot}_{ij} = sig^{'}_{ij} + p*\delta_{ij}
  for (i=0;i<3;i++)
    sigt[i][i] += pp;
  
  // convert total stress tensor back to vector form - sig
  tensor_vector(sig, sigt, Mm->ip[ipp].ssst,stress);
  
  // store the total stresses 
  for (i=0;i<ncomp;i++)
    Mm->ip[ipp].stress[i] = sig[i];    
}


/**
  The funtion marks required non-mechanical quantities in the array anmq.

  @param anmq - array with flags for used material types
                anmq[i] = 1 => qunatity type nonmechquant(i+1) is required
                anmq[i] = 0 => qunatity type nonmechquant(i+1) is not required

  @return The function does not return anything, but it may change content of anmq array.
*/
void effstress::give_reqnmq(long *anmq)
{
  switch (ppct)
  {
    case noporep:
      break;
    case const_all:
    case nonmechq:
      anmq[pore_pressure-1] = 1;
      break;
    default:
      print_err("Unknown type of pore pressure computation is required", __FILE__, __LINE__, __func__);
  }
}


/**
  The function initializes corresponding component of the mechmat::nonmechq array
  in the case of constant pore pressure value defined in the material.
  Additionally, it sets flag probdesc::eigstrcomp for computation of nodal force 
  contribution due to pore pressure.

  @param ipp - integration point pointer

  @return The function does not return anything but it changes Mm->nonmechq array.

  Created by Tomas Koudelka, 14.10.2013 
*/
void effstress::initvalues(long ipp)
{
  if (ppct == const_all)
  {
    // set pore_pressure_component of nonmechq array to constant value
    Mm->storenonmechq(pore_pressure, ipp, ppv);   
  }
}
